(use ugarit-core)
(use miscmacros)
(use ugarit-backend)
(use ugarit-streams)
(use ugarit-files)
(use ugarit-snapshot)
(use ugarit-archive)
(use ugarit-vfs)
(use ugarit-api)
(use test)
(use posix)
(use posix-extras)
(use directory-rules)
(use matchable)
(use sql-de-lite)
(include "../backend-devtools.scm")

;; Test egg extensions

(define-syntax test-no-errors
  (syntax-rules ()
    ((_ expr)
     (test-no-errors (->string expr) expr))
    ((_ name expr)
     (test name (void) (begin expr (void))))))

(define-syntax test-define
  (syntax-rules ()
    ((_ var expr)
     (test-define (->string '(define var expr)) var expr))
    ((_ name var expr)
     (begin
       (define var (void))
       (test-no-errors name (set! var expr))))))

(define-syntax test-define-values
  (syntax-rules ()
    ((_ (var ...) expr)
     (test-define-values (->string '(define-values (var ...) expr)) (var ...) expr))
    ((_ name (var ...) expr)
     (begin
       (define var (void)) ...
       (test-no-errors name (set!-values (var ...) expr))))))

;; Test utilities

; Basic smoke test of a backend
(define (test-backend w)
  (parameterize
   ((backend-log! (lambda (type message) (void))))
   (test-assert "Storage block size is valid" (exact? (storage-max-block-size w))) 
   (test-assert "Storage writable" (storage-writable? w))
   (test-assert "Storage is empty" (not ((storage-exists? w) "TEST")))

   (test "Load a block" (void) ((storage-put! w) "TEST" (list->u8vector (list 1 2 3 4 5)) 'test))
   (test-assert "Block successfully loaded" ((storage-exists? w) "TEST"))
   (test "Block contents reads back" (list 1 2 3 4 5) (u8vector->list ((storage-get w) "TEST")))

   (let ((long-name "caeef4a6ffe0cc5e25f9966c922366ec36b2bcf0dcd40754991ffe107b49fb33"))

     (test "Nonexistant block with a long name reacts correctly" #f ((storage-get w) long-name))
     (test "Load a block with a long name" (void) ((storage-put! w) long-name (list->u8vector (list 6 7 8 9 10)) 'test))
     (test-assert "Block with a long name successfully loaded" ((storage-exists? w) long-name))
     (test "Block contents with a long name reads back" (list 6 7 8 9 10) (u8vector->list ((storage-get w) long-name))))

   (test "Nonexistant block reacts correctly" #f ((storage-get w) "NONEXISTANT"))
   (test-error "Cannot update existing blocks" ((storage-put! w) "TEST" (list->u8vector (list 1 2 3 4 5 6)) 'test))
   (if (storage-unlinkable? w)
       (begin
         (test "Unlink returns data" (list 1 2 3 4 5) (u8vector->list ((storage-unlink! w) "TEST")))
         (test-assert "Unlinked block is gone" (not ((storage-exists? w) "TEST")))))

   (test "Set a tag" (void) ((storage-set-tag! w) "TEST" "TEST123"))
   (test "Tag is not locked" #f ((storage-tag-locked? w) "TEST"))
   (test "Lock a tag" #t ((storage-lock-tag! w) "TEST"))
   (test "Tag is now locked" #t ((storage-tag-locked? w) "TEST"))
   (test "Lock a tag again" #f ((storage-lock-tag! w) "TEST"))
   (test "Tag is still locked" #t ((storage-tag-locked? w) "TEST"))
   (test "Unlock a tag" (void) ((storage-unlock-tag! w) "TEST"))
   (test "Tag is no longer locked" #f ((storage-tag-locked? w) "TEST"))
   (test "Tag reads back" "TEST123" ((storage-tag w) "TEST"))
   (test "Tag list works" (list "TEST") ((storage-all-tags w)))
   (test "Remove tag" (void) ((storage-remove-tag! w) "TEST"))

   (test "Nonexistant tag is not locked" #f ((storage-tag-locked? w) "NONEXISTANT"))
   (test "Lock a nonexistant tag" #t ((storage-lock-tag! w) "NONEXISTANT"))
   (test "Nonexistant tag is now locked" #t ((storage-tag-locked? w) "NONEXISTANT"))
   (test "Lock a nonexistant tag again" #f ((storage-lock-tag! w) "NONEXISTANT"))
   (test "Nonexistant tag is still locked" #t ((storage-tag-locked? w) "NONEXISTANT"))
   (test "Unlock a locked nonexistant tag" (void) ((storage-unlock-tag! w) "NONEXISTANT"))
   (test "Nonexistant tag is no longer locked" #f ((storage-tag-locked? w) "NONEXISTANT"))

   (test "Close storage" (void) ((storage-close! w)))))

(define (key-stream-cat a ks-hash ks-type level)
   (define type (vault-exists? a ks-hash))
   (if (eq? type ks-type)
      (begin
         (printf "ks(~A): ~A (~A)\n" level ks-hash type)
            (for-each (lambda (subkey)
               (key-stream-cat a subkey ks-type (+ level 1)))
               (deserialise-key-stream (vault-get a ks-hash type))))
      (printf "kleaf(~A): ~A (~A)\n" level ks-hash type)))

(define (sexpr-stream-cat a ss-hash leaf-type ss-type level)
   (define type (vault-exists? a ss-hash))
   (test-assert "sexpr stream internal consistency" (or (eq? type ss-type) (eq? type leaf-type)))
   (cond
      ((eq? type ss-type)
         (begin ; key stream node
            (printf "ss(~A): ~A (~A)\n" level ss-hash type)
               (for-each (lambda (subkey)
                  (sexpr-stream-cat a subkey leaf-type ss-type (+ level 1)))
                  (deserialise-key-stream (vault-get a ss-hash type)))))
      ((eq? type leaf-type)
         (begin ; leaf node
            (printf "sleaf(~A): ~A (~A)\n" level ss-hash type)
               (for-each (lambda (sexpr)
                  (printf " ~A\n" sexpr))
                  (deserialise-sexpr-stream (vault-get a ss-hash type)))))))

(define (check-vault-is-empty a db-path)
  ;; We can check if sqlite vaults are empty, by peering inside!
  ;; FIXME: Half the time, the db is locked!
  ;; (when db-path
  ;;  (let* ((db (open-database db-path))
  ;;         (blocks (car (query fetch (sql db "SELECT COUNT(*) FROM blocks WHERE type <> 'ugarit-vault-configuration';"))))
  ;;         (header-blocks (car (query fetch (sql db "SELECT COUNT(*) FROM blocks WHERE type = 'ugarit-vault-configuration';"))))
  ;;         (tags (car (query fetch (sql db "SELECT COUNT(*) FROM tags WHERE tag <> '#ugarit-vault-configuration';"))))
  ;;         (header-tags (car (query fetch (sql db "SELECT COUNT(*) FROM tags WHERE tag = '#ugarit-vault-configuration';")))))
  ;;    (test "Vault has no blocks other than the header" 0 blocks)
  ;;    (test "Vault has one header block" 1 header-blocks)
  ;;    (test "Vault has no tags other than the header" 0 tags)
  ;;    (test "Vault has one header tag" 1 header-tags)
  ;;    (close-database db)))
  (void))

(define (check-extract-results path plain-file1-contents plain-file2-contents)
  (let* ((tp (lambda (relative) (string-append path "/" relative)))
         (check-file (lambda (relative contents)
                       (when contents
                             (test (sprintf "Contents of ~s are as expected" relative)
                                   contents
                                   (with-input-from-file (tp relative) read))))))

    ;; Plain file(s)
    (check-file "plain-file1" plain-file1-contents)
    (check-file "plain-file2" plain-file2-contents)

    ;; FIFO
    (test-assert "FIFO exists" (fifo? (tp "fifo")))

    ;; Specials
    (if (zero? (current-user-id))
        (begin
          (let* ((stats (file-stat (tp "block-special")))
                 (type (bitwise-and (vector-ref stats 1) stat/ifmt))
                 (devnum (vector-ref stats 10)))
            (test-assert "Block special file exists" (eq? type stat/ifblk))
            (test "Block special file has correct devnum" 123 devnum))

          (let* ((stats (file-stat (tp "character-special")))
                 (type (bitwise-and (vector-ref stats 1) stat/ifmt))
                 (devnum (vector-ref stats 10)))
            (test-assert "Character special file exists" (eq? type stat/ifchr))
            (test "Character special file has correct devnum" 456 devnum))))

    ;; Directory
    (test-assert "Directory exists" (directory? (tp "directory")))))

(define (dirent->sexpr dirent)
  (cons (dirent-subnode-key dirent)
        (cons (dirent-name dirent)
              (cons (dirent-type dirent)
                    (dirent-props dirent)))))

(define (dirent-list dirent acc)
  (cons (dirent->sexpr dirent) acc))

;; Vault operations smoke test

(define (test-vault-basics a store-path #!optional test-key)
  (test-group "Basic vault operations"
              (test-assert "Vault max block size is valid"
                           (exact? (vault-max-block-size a)))
              (define test-list (list 1 2 3 4 5))
              (define test-data (list->u8vector test-list))
              (unless test-key
                      (printf "Test key should be ~A\n" ((vault-hash a) test-data 'test)))
              (test "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-assert "Vault is writable" (vault-writable? a))
              (test-assert "Key does not already exist" (not (vault-exists? a test-key)))
              (test "Data goes into vault" (void) (vault-put! a test-key test-data 'test))
              (test-assert "Data now exists in vault" (vault-exists? a test-key))
              (test "Data reads back" test-list (u8vector->list (vault-get a test-key 'test)))

              (if (vault-unlinkable? a)
                  (begin
                    (test "Unlink returns data" test-list (u8vector->list (vault-unlink! a test-key)))
                    (test-assert "Unlinked data is gone" (not (vault-exists? a test-key)))))
              (test "Tag setting" (void) (vault-set-tag! a (make-tag "TEST" 'test test-key)))

              (test "Tag is not locked" #f (vault-tag-locked? a "TEST"))
              (test "Lock a tag" #t (vault-lock-tag! a "TEST"))
              (test "Tag is now locked" #t (vault-tag-locked? a "TEST"))
              (test "Tag is still locked" #t (vault-tag-locked? a "TEST"))
              (test "Unlock a tag" (void) (vault-unlock-tag! a "TEST"))
              (test "Tag is no longer locked" #f (vault-tag-locked? a "TEST"))

              (test-define "Tag reading" tag-body (vault-tag a "TEST"))
              (test "Tag reading - name" "TEST" (tag-name tag-body))
              (test "Tag reading - type" 'test (tag-type tag-body))
              (test "Tag reading - key" test-key (tag-key tag-body))

              (test "Tag listing 1" (list "TEST") (vault-all-tags a))
              (test "Tag listing 2" (list "#ugarit-vault-configuration" "TEST")
                    (sort (vault-all-tags a #t) string<?))
              (test "Tag listing 3" (list "TEST") (vault-all-tags a #f))

              (test "System tag setting" (void) (vault-set-tag! a (make-tag "#SYSTEM" 'test test-key)))

              (test "Tag listing 4" (list "TEST") (vault-all-tags a))
              (test "Tag listing 5" (list "#SYSTEM" "#ugarit-vault-configuration" "TEST")
                    (sort (vault-all-tags a #t) string<?))
              (test "Tag listing 6" (list "TEST") (vault-all-tags a #f))

              (test "Tag removal 1" (void) (vault-remove-tag! a "TEST"))
              (test "Tag removal 2" (void) (vault-remove-tag! a "#SYSTEM"))

              (test "No tag listing 1" (list) (vault-all-tags a))
              (test "No tag listing 2" (list "#ugarit-vault-configuration") (vault-all-tags a #t))
              (test "No tag listing 3" (list) (vault-all-tags a #f))

              (test "Vault format version" 1 (vault-format-version a))

              (if (vault-unlinkable? a)
                  (check-vault-is-empty a store-path))))

;; Test key/sexpr-streams
(define (test-vault-streams a store-path)
  (test-group "0-element key stream"
              (define test-list (list 1 2 3 4 5 6))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-define "Create key-stream writer" ksw (make-key-stream-writer* a 'test-ks))
              (test-define-values "Close key-stream writer" (ks-hash ks-reused?) ((key-stream-writer-finish! ksw)))
              (test-assert "Key stream did not already exist" (not ks-reused?))
              (test "Key stream reads back OK" '() (fold-key-stream a ks-hash 'test-ks cons '()))

              (if (vault-unlinkable? a)
                  (begin
                    (test-assert (vault-unlink! a ks-hash))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "1-element key stream...\n"
              (define test-list (list 1 2 3 4 5 6 7))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-define "Create key-stream writer" ksw (make-key-stream-writer* a 'test-ks))
              (test "Store a block into the vault" (void) (vault-put! a test-key test-data 'test))
              (test "Insert hash into key stream" (void) ((key-stream-writer-write! ksw) test-key #f))
              (test-define-values "Close key-stream writer" (ks-hash ks-reused?) ((key-stream-writer-finish! ksw)))
              (test-assert "Key stream did not already exist" (not ks-reused?))
              (test "Correct hash" test-key ks-hash)

              (test "Correct result from reading back key stream"
                    (list (cons test-key 'test))
                    (fold-key-stream a ks-hash 'test-ks
                                     (lambda (key type acc)
                                       (cons (cons key type) acc)) '()))

              (if (vault-unlinkable? a)
                  (begin
                    (test-assert (vault-unlink! a test-key))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "2-element key stream...\n"
              (define test-list (list 1 2 3 4 5 6 7 8))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-define "Create key-stream writer" ksw (make-key-stream-writer* a 'test-ks))
              (test "Store a block into the vault" (void) (vault-put! a test-key test-data 'test))
              (test "Insert hash 1 into key stream" (void) ((key-stream-writer-write! ksw) test-key #f))
              (test "Insert hash 2 into key stream" (void) ((key-stream-writer-write! ksw) test-key #t))
              (test-define-values "Close key-stream writer" (ks-hash ks-reused?) ((key-stream-writer-finish! ksw)))
              (test-assert "Key stream did not already exist" (not ks-reused?))
              (test-assert "Key stream now exists" (vault-exists? a ks-hash))

              (test "Correct result from reading back key stream"
                    (list (cons test-key 'test) (cons test-key 'test))
                    (fold-key-stream a ks-hash 'test-ks
                                     (lambda (key type acc)
                                       (cons (cons key type) acc)) '()))

              (if (vault-unlinkable? a)
                  (begin
                    (test-assert (vault-unlink! a ks-hash))
                    (test-assert (not (vault-unlink! a test-key)))
                    (test-assert (vault-unlink! a test-key))
                    (check-vault-is-empty a store-path)
                    )))

  (define iterations 1024)
  (test-group (sprintf "~a-element key stream..." iterations)
              (define test-list (list 1 2 3 4 5 6 7 8 9))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-define "Create key-stream writer" ksw (make-key-stream-writer* a 'test-ks))
              (test "Store a block into the vault" (void) (vault-put! a test-key test-data 'test))
              (test "Insert hash 1 into key stream" (void) ((key-stream-writer-write! ksw) test-key #t)) ; Ensure one reference is left at the end
              (test "Insert more hashes into key stream" (void)
                    (dotimes (iter iterations)
                             ((key-stream-writer-write! ksw) test-key #t)))
              (test-define-values "Close key-stream writer" (ks-hash ks-reused?) ((key-stream-writer-finish! ksw)))
              (test-assert "Key stream did not already exist" (not ks-reused?))
              (test-assert "Key stream now exists" (vault-exists? a ks-hash))

              (test-define "Key stream reads back OK" keys
                           (fold-key-stream a ks-hash 'test-ks
                                            (lambda (key type acc)
                                              (cons (cons key type) acc)) '()))
              (test "Correct number of keys come back" (+ 1 iterations) (length keys))
              (test-assert "Correct keys come back" (every
                                                     (lambda (key) (string=? (car key) test-key))
                                                     keys))

              (if (vault-unlinkable? a)
                  (begin
                    (test "Unlink key stream" (void)
                          (unlink-key-stream! a ks-hash 'test-ks
                                              (lambda (vault key type)
                                                (test-assert "Unlink non-final block" (not (vault-unlink! vault key))))))
                    (test-assert "Unlink final block" (vault-unlink! a test-key))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group (sprintf "~a-element key stream ordering..." iterations)
              (test-define "Create key-stream writer" ksw (make-key-stream-writer* a 'test-ks))
              (test "Insert hashes into key stream" (void)
                    (dotimes (iter iterations)
                             (let* ((test-block
                                     (blob->u8vector (string->blob
                                                     (string-append "test-data"
                                                                    (number->string iter)))))
                                    (test-type (string->symbol
                                                (number->string iter)))
                                    (test-key ((vault-hash a) test-block test-type)))
                               (vault-put! a test-key test-block
                                           test-type)
                               ((key-stream-writer-write! ksw) test-key #f))))
              (test-define-values "Close key-stream writer" (ks-hash ks-reused?) ((key-stream-writer-finish! ksw)))
              (test-assert "Key stream did not already exist" (not ks-reused?))
              (test-assert "Key stream now exists" (vault-exists? a ks-hash))

              (test-define "Key stream reads back OK" keys
                           (reverse
                            (fold-key-stream a ks-hash 'test-ks
                                             (lambda (key type acc)
                                               (cons (cons key type) acc)) '())))
              (test "Correct number of keys come back" iterations (length keys))

              (test "Keys come back in correct order"
                    ;; Use the fact the block types are encoded integers
                    (iota iterations)
                    (map (lambda (ent)
                         (string->number
                          (symbol->string
                           (cdr ent)))) keys))

              (if (vault-unlinkable? a)
                  (begin
                    (test "Unlink key stream" (void)
                          (unlink-key-stream! a ks-hash 'test-ks
                                              (lambda (vault key type)
                                                (test-assert "Unlink block" (vault-unlink! vault key)))))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "0-element sexpr stream"
              (test-define "Create sexpr-stream writer" ssw (make-sexpr-stream-writer* a 't 'ti))
              (test-define-values "Close sexpr-stream writer" (ss-hash ss-reused?) ((sexpr-stream-writer-finish! ssw)))
              (test-assert "Sexpr stream did not already exist" (not ss-reused?))
              (test "Sexpr stream reads back OK" '() (fold-sexpr-stream a ss-hash 't 'ti cons '()))

              (if (vault-unlinkable? a)
                  (begin
                    (test-assert (vault-unlink! a ss-hash))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "1-element sexpr stream"
              (define test-list (list 1 2 3 4 5 6 7 8 9 10 11))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test-define "Create sexpr-stream writer" ssw (make-sexpr-stream-writer* a 't 'ti))
              (test "Write to sexpr stream" (void)
                    ((sexpr-stream-writer-write! ssw) `(foo ,test-key) (list (cons test-key #f))))
              (test-define-values "Close sexpr-stream writer" (ss-hash ss-reused?) ((sexpr-stream-writer-finish! ssw)))
              (test-assert "Sexpr stream did not already exist" (not ss-reused?))
              (test "Sexpr stream reads back OK"
                    `((foo ,test-key))
                    (fold-sexpr-stream a ss-hash 't 'ti cons '()))

              (if (vault-unlinkable? a)
                  (begin
                    (test "Unlink sexpr stream" (void)
                          (unlink-sexpr-stream! a ss-hash 't 'ti
                                                (lambda (sexpr)
                                                  (test "Correct entry read back" `(foo ,test-key) (identity sexpr))
                                                  (test "Unlink entry" #f (vault-unlink! a test-key)))))
                    (test-assert "Sexpr stream is gone" (not (vault-exists? a ss-hash)))
                    (test-assert "Test block is gone" (not (vault-exists? a test-key)))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "2-element sexpr stream"
              (define test-list (list 1 2 3 4 5 6 7 8 9 10 11 12))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test "Vault write" (void) (vault-put! a test-key test-data 'test))
              (test-define "Create sexpr-stream writer" ssw (make-sexpr-stream-writer* a 't 'ti))
              (test "Write to sexpr stream" (void)
                    ((sexpr-stream-writer-write! ssw) `(foo ,test-key) (list (cons test-key #f))))
              (test "Write to sexpr stream" (void)
                    ((sexpr-stream-writer-write! ssw) `(foo ,test-key) (list (cons test-key #t))))
              (test-define-values "Close sexpr-stream writer" (ss-hash ss-reused?) ((sexpr-stream-writer-finish! ssw)))
              (test-assert "Sexpr stream did not already exist" (not ss-reused?))
              (test "Sexpr stream reads back OK"
                    `((foo ,test-key) (foo ,test-key))
                    (fold-sexpr-stream a ss-hash 't 'ti cons '()))

              (if (vault-unlinkable? a)
                  (begin
                    (define unlinks 0)

                    (test "Unlink sexpr stream" (void)
                          (unlink-sexpr-stream! a ss-hash 't 'ti
                                                (lambda (sexpr)
                                                  (test "Correct entry read back" `(foo ,test-key) (identity sexpr))
                                                  (test (sprintf "Unlink entry ~a" unlinks) (if (zero? unlinks) #f test-data) (vault-unlink! a test-key))
                                                  (set! unlinks (+ unlinks 1)))))
                    (test-assert "Sexpr stream is gone" (not (vault-exists? a ss-hash)))
                    (test-assert "Test block is gone" (not (vault-exists? a test-key)))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group (sprintf "~A-element sexpr stream" iterations)
              (define test-list (list 1 2 3 4 5 6 7 8 9 10 11 12 13))
              (define test-data (list->u8vector test-list))
              (test-define "Vault hash" test-key ((vault-hash a) test-data 'test))
              (test "Vault write" (void) (vault-put! a test-key test-data 'test))
              (test-define "Create sexpr-stream writer" ssw (make-sexpr-stream-writer* a 't 'ti))
              ((sexpr-stream-writer-write! ssw) `(foo ,test-key) (list (cons test-key #f)))
              (dotimes (iter iterations)
                       (test "Write to sexpr stream" (void)
                             ((sexpr-stream-writer-write! ssw) `(foo ,test-key) (list (cons test-key #t)))))
              (test-define-values "Close sexpr-stream writer" (ss-hash ss-reused?) ((sexpr-stream-writer-finish! ssw)))
              (test-assert "Sexpr stream did not already exist" (not ss-reused?))
              (test-define "Sexpr stream reads back OK" sexprs
                           (fold-sexpr-stream a ss-hash 't 'ti cons '()))
              (test "Correct number of sexprs read back" (+ 1 iterations) (length sexprs))
              (test-assert "Correct sexprs read back"
                           (every
                            (lambda (sexpr) (equal? sexpr `(foo ,test-key)))
                            sexprs))

              (if (vault-unlinkable? a)
                  (begin
                    (define unlinks 0)

                    (test "Unlink sexpr stream" (void)
                          (unlink-sexpr-stream! a ss-hash 't 'ti
                                                (lambda (sexpr)
                                                  (test "Correct entry read back" `(foo ,test-key) (identity sexpr))
                                                  (test (sprintf "Unlink entry ~a" unlinks) (if (< unlinks iterations) #f test-data) (vault-unlink! a test-key))
                                                  (set! unlinks (+ unlinks 1)))))
                    (test-assert "Sexpr stream is gone" (not (vault-exists? a ss-hash)))
                    (test-assert "Test block is gone" (not (vault-exists? a test-key)))
                    (check-vault-is-empty a store-path)))))

;; Test file and directory storage
(define (test-vault-files a store-path)
  (test-group "Files"
              (define test-string "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.")

              (with-output-to-file "./test-file"
                (lambda ()
                  (display test-string)))

              (test-define-values "Store a file" (file-key file-reused?)
                                  (store-file! a "./test-file" (vector 0 0 0 0 0 0 0 0 0 0 0 0 0)))


              (test "Read it back" test-string
                    (let ((fd (file-open "./test-file-2" (+ open/wronly open/creat))))
                      (write-file-contents a file-key fd)
                      (file-close fd)
                      (with-input-from-file "./test-file-2"
                        (lambda ()
                          (blob->string
                           (u8vector->blob/shared
                            (read-u8vector #f)))))))

              (if (vault-unlinkable? a)
                  (begin
                    (test "Delete the file" (void) (unlink-file! a file-key))
                    (check-vault-is-empty a store-path)
                    )))

  (test-group "Directories"
              (let* ((test-dir (string-append store-path "-test-data"))
                     (extract1-dir (string-append store-path "-test-extract1"))
                     (extract2-dir (string-append store-path "-test-extract2"))
                     (extract3-dir (string-append store-path "-test-extract3"))
                     (extract4-dir (string-append store-path "-test-extract4"))
                     (tp (lambda (relative)
                           (string-append test-dir "/" relative))))
                (create-directory test-dir)
                (with-output-to-file (tp "plain-file1") (lambda () (write "Hello world")))
                (create-fifo (tp "fifo"))
                ;; These two need root!
                (if (zero? (current-user-id))
                    (begin
                      (create-special-file (tp "block-special") stat/ifblk 123)
                      (create-special-file (tp "character-special") stat/ifchr 456)))
                (create-directory (tp "directory"))

                ;; Dump it
                (test-define-values "Store a directory" (dir1-key
                                                         dir1-reused?
                                                         files bytes)
                                    (call-with-context-support
                                     (vault-global-directory-rules a)
                                     (lambda ()
                                       (store-directory! a test-dir))))

                (create-directory extract1-dir)

                                        ; FIXME: Set up a progress callback that counts the
                                        ; bytes and files and checks they match the
                                        ; directory metadata in the vault on every extract
                (test "Extract a directory" (void)
                      (extract-directory! a dir1-key extract1-dir))

                (check-extract-results extract1-dir "Hello world" #f)

                ;; Now add an extra file and dump again
                (with-output-to-file (tp "plain-file2") (lambda () (write "Hello world 2")))

                (test-define-values "Store a directory again"
                                    (dir2-key dir2-reused?
                                              dir2-files dir2-bytes)
                                    (call-with-context-support
                                     (vault-global-directory-rules a)
                                     (lambda ()
                                       (store-directory! a test-dir))))

                (test-assert "Changed directory is not reused" (not dir2-reused?))

                (create-directory extract2-dir)
                (test "Extract a directory" (void)
                      (extract-directory! a dir2-key extract2-dir))

                (check-extract-results extract2-dir "Hello world" "Hello world 2")

                ;; Now change an existing file and dump again
                (with-output-to-file (tp "plain-file1") (lambda () (write "Hello world again!")))

                (test-define-values "Store a directory again"
                                    (dir3-key dir3-reused?
                                              dir3-files dir3-bytes)
                                    (call-with-context-support
                                     (vault-global-directory-rules a)
                                     (lambda ()
                                       (store-directory! a test-dir))))

                (test-assert "Changed directory is not reused" (not dir3-reused?))

                (create-directory extract3-dir)
                (test "Extract a directory" (void)
                      (extract-directory! a dir3-key extract3-dir))

                (check-extract-results extract3-dir "Hello world again!" "Hello world 2")

                ;; Now make no changes and dump again
                (test-define-values "Store a directory again"
                                    (dir4-key dir4-reused?
                                              dir4-files dir4-bytes)
                                    (call-with-context-support
                                     (vault-global-directory-rules a)
                                     (lambda ()
                                       (store-directory! a test-dir))))

                (test-assert "Unchanged directory is reused" dir4-reused?)
                (test "Mark reused directory" (void) (vault-link! a dir4-key))
                (create-directory extract4-dir)
                (test "Extract a directory" (void)
                      (extract-directory! a dir4-key extract4-dir))
                (check-extract-results extract4-dir "Hello world again!" "Hello world 2")

                ;; Tidy up
                (when (vault-unlinkable? a)
                      (begin
                        (test "Delete the first directory" (void) (unlink-directory! a dir1-key))
                        (check-extract-results extract2-dir "Hello world" "Hello world 2")
                        (test "Delete the second directory" (void) (unlink-directory! a dir2-key))
                        (check-extract-results extract3-dir "Hello world again!" "Hello world 2")
                        (test "Delete the third directory" (void) (unlink-directory! a dir3-key))
                        (check-extract-results extract4-dir "Hello world again!" "Hello world 2")
                        (test "Delete the fourth directory" (void) (unlink-directory! a dir4-key))
                        (check-vault-is-empty a store-path))))))

;; Test snapshot algorithms
(define (test-vault-snapshots a store-path)
  (test-group
   "Snapshots"
   (let ((test-job (make-job #f #t)))
     (parameterize
      ((current-job test-job))
      (let* ((test-dir (string-append store-path "-test-data")))
        (test-define-values "Store a directory"
                            (dir-key dir-reused?
                                     dir-files dir-bytes)
                            (call-with-context-support
                             (vault-global-directory-rules a)
                             (lambda ()
                               (store-directory! a test-dir))))

        (if (vault-unlinkable? a)
            (test-assert "Directory was not reused" (not dir-reused?)))

        (test-define-values "Tag it as a snapshot" (sk1)
                            (tag-snapshot! a "Test" dir-key dir-reused? (list) test-job))

        (test-define-values "Store the directory again"
                            (dir2-key dir2-reused?
                                      dir2-files dir2-bytes)
                            (call-with-context-support
                             (vault-global-directory-rules a)
                             (lambda ()
                               (store-directory! a test-dir))))

        (test-assert "Directory was reused" dir2-reused?)

        (test-assert "Directory has the same key" (string=? dir-key dir2-key))
        (test "Log a message" (void) (job-log! 'info #f "This is a test"))

        (test-define-values "Tag it as a snapshot" (sk2)
                            (tag-snapshot! a "Test" dir2-key dir2-reused? (list) test-job))

        (test-define-values "Read the tag back" (tag2) (tag-key (vault-tag a "Test")))
        (test-define-values "Walk the history with fold-history" (result)
                            (fold-history a "Test" tag2 'snapshot
                                          (lambda (snapshot-key name snapshot acc)
                                            (cons snapshot acc))
                                          '()))

        (test-assert "History has expected form"
                     (match result
                            (((('mtime . _)
                               ('contents . dir-key*)
                               ('stats . _)
                               ('log . log-key1))
                              (('previous . sk1*)
                               ('mtime . _)
                               ('contents . dir2-key*)
                               ('stats . _)
                               ('log . log-key2)))
                             (and (string=? sk1 sk1*)
                                  (string=? dir2-key dir2-key*)
                                  (string=? dir-key dir-key*)
                                  (string? log-key1)
                                  (string? log-key2)
                                  (equal?
                                   '()
                                   (fold-sexpr-stream a log-key1 'f 'fi cons '())
                                   )
                                  (match
                                   (fold-sexpr-stream a log-key2 'f 'fi cons '())
                                   ((('info _ #f "This is a test")) #t)
                                   (else #f))))
                            (else #f)))

        (test "Walk the tag list with fold-vault-node"
              `(((tag . "Test")
                 "Test"
                 tag
                 (current . ,sk2)
                 (type . snapshot)
                 (locked . #f)))
              (fold-vault-node a '() dirent-list '()))

        (test-define-values "Walk the history of tag 'Test' with fold-vault-node" (tag)
                            (fold-vault-node a '(tag . "Test") dirent-list '()))

        (test-assert "Tag history has expected form"
                     (match tag
                            (((sk-c***
                               _
                               'snapshot
                               ('mtime . _)
                               ('contents . dir-key***)
                               ('stats . _)
                               ('log . log-key1))
                              (sk-c**
                               _
                               'snapshot
                               ('previous . sk1**)
                               ('mtime . _)
                               ('contents . dir-key**)
                               ('stats . _)
                               ('log . log-key2))
                              (sk-c*
                               "current"
                               'snapshot
                               ('previous . sk1*)
                               ('mtime . _)
                               ('contents . dir-key*)
                               ('stats . _)
                               ('log . log-key3)))
                             (and
                              (string=? sk1 sk1*)
                              (string=? sk2 sk-c*)
                              (string=? dir-key dir-key*)
                              (string=? sk1 sk1**)
                              (string=? sk2 sk-c**)
                              (string=? dir-key dir-key**)
                              (string=? sk1 sk-c***)
                              (string=? dir-key dir-key***)
                              (string? log-key1)
                              (string? log-key2)
                              (string? log-key3)
                              (equal?
                               '()
                               (fold-sexpr-stream a log-key1 'f 'fi cons '()))
                              (match
                               (fold-sexpr-stream a log-key2 'f 'fi cons '())
                               ((('info _ #f "This is a test")) #t)
                               (else #f))
                              (match
                               (fold-sexpr-stream a log-key3 'f 'fi cons '())
                               ((('info _ #f "This is a test")) #t)
                               (else #f))))
                            (else #f)))

        (test-define-values "Walk the snapshot metadata directory with fold-vault-node" (snapshot-meta)
                            (fold-vault-node a sk1 dirent-list '()))

        (test-assert "Snapshot metadata has the expected form"
                     (match snapshot-meta
                            (((dir-key*
                               "contents"
                               'dir
                               ('mtime . _)
                               ('contents . dir-key**)
                               ('stats . _)
                               ('log . log-key))
                              (#f
                               "properties.sexpr"
                               'inline
                               ('text . _)
                               (mode . 256))
                              (#f
                               "log.sexpr"
                               'file
                               ('contents . log-key)
                               (mode . 256)))
                             (and
                              (string=? dir-key dir-key*)
                              (string=? dir-key dir-key**)))
                            (else #f)))


        (test-define-values "Walk the root directory with fold-vault-node" (dir)
                            (fold-vault-node a dir-key dirent-list '()))
        (if (zero? (current-user-id))
            (test-assert "Directory listing has the expected form (as root)"
                         (match dir
                                (((#f "block-special" 'block-device (number . 123) . _)
                                  (#f "character-special" 'character-device (number . 456) . _)
                                  (_ "directory" 'dir . _)
                                  (#f "fifo" 'fifo . _)
                                  (#f "plain-file1" 'file . _)
                                  (#f "plain-file2" 'file . _)) #t)
                                (else #f)))

            (test-assert "Directory listing has the expected form (not as root)"
                         (match dir
                                (((_ "directory" 'dir . _)
                                  (#f "fifo" 'fifo . _)
                                  (#f "plain-file1" 'file . _)
                                  (#f "plain-file2" 'file . _)) #t)
                                (else #f))))

        (test-group "Path traversals"
                    (test "()" '(() "/" root) (dirent->sexpr (traverse-vault-path a '())))
                    (test "(.)" '(() "/" root) (dirent->sexpr (traverse-vault-path a '("."))))
                    (test-assert "(\"Test\")" 
                                 (match (dirent->sexpr (traverse-vault-path a '("Test")))
                                        ((('tag . "Test")
                                          "Test"
                                          'tag
                                          ('current . sk2*)
                                          ('type . 'snapshot)
                                          ('locked . #f))
                                         (and
                                          (string=? sk2 sk2*)))
                                        (else #f)))
                    (test-assert "(\"Test\" .)" 
                                 (match (dirent->sexpr (traverse-vault-path a '("Test" ".")))
                                        ((('tag . "Test")
                                          "Test"
                                          'tag
                                          ('current . sk2*)
                                          ('type . 'snapshot)
                                          ('locked . #f))
                                         (and
                                          (string=? sk2 sk2*)))
                                        (else #f)))
                    (test "(\"Test\" ..)" '(() "/" root) (dirent->sexpr (traverse-vault-path a '("Test" ".."))))
                    (test-assert "(\"Test\" \"current\")"
                                 (match (dirent->sexpr (traverse-vault-path a '("Test" "current")))
                                        ((sk2*
                                          "current"
                                          'snapshot
                                          ('previous . sk1*)
                                          ('mtime . _)
                                          ('contents . dk2*)
                                          ('stats . _)
                                          ('log . _))
                                         (and
                                          (string=? sk1 sk1*)
                                          (string=? sk2 sk2*)
                                          (string=? dir-key dk2*)))
                                        (else #f)))
                    (test-assert "(\"Test\" \"current\" \"contents\")"
                                 (match (dirent->sexpr (traverse-vault-path a '("Test" "current" "contents")))
                                        ((dk2*
                                          "contents"
                                          'dir
                                          ('previous . sk1*)
                                          ('mtime . _)
                                          ('contents . dk2**)
                                          ('stats . _)
                                          ('log . _))
                                         (and
                                          (string=? sk1 sk1*)
                                          (string=? dir-key dk2*)
                                          (string=? dir-key dk2**)))
                                        (else #f)))
                    (test-assert "(\"Test\" \"current\" \"contents\" \"directory\")"
                                 (match (dirent->sexpr (traverse-vault-path a '("Test" "current" "contents" "directory")))
                                        ((sdk2*
                                          "directory"
                                          'dir
                                          ('contents . sdk2**)
                                          ('files . 0)
                                          ('size . 0)
                                          ('mode . 493)
                                          ('uid . _)
                                          ('gid . _)
                                          ('mtime . _))
                                         #t)
                                        (else #f)))
                    (test-assert "(\"Test\" \"current\" \"contents\" \"plain-file1\")"
                                 (match (dirent->sexpr (traverse-vault-path a '("Test" "current" "contents" "plain-file1")))
                                        ((#f
                                          "plain-file1"
                                          'file
                                          ('contents . pfk1*)
                                          ('size . 20)
                                          ('mode . 420)
                                          ('uid . _)
                                          ('gid . _)
                                          ('mtime . _))
                                         #t)
                                        (else #f)))))))

   ;; FIXME [a987e28fef] Fork the snapshot chain, then merge it again
   ;; and see how it folds.
))

;; Test archives
(define (test-vault-archives a store-path)
  (test-group
   "Archives"
   (let ((test-job (make-job #f #t)))
     (parameterize
      ((current-job test-job))
      (let ((test-file-1 (string-append store-path "-arch1"))
            (test-file-2 (string-append store-path "-arch2")))
        (with-output-to-file test-file-1
          (lambda () (display "This is the contents of archive file 1")))
        (with-output-to-file test-file-2
          (lambda () (display "This is the contents of archive file 2")))

        (test-define-values "Store a file 1" (file1-key file1-reused?)
                            (store-file! a test-file-1 (vector 0 0 0 0 0 0 0 0 0 0 0 0 0)))

        (test-define-values "Store a file 2" (file2-key file2-reused?)
                            (store-file! a test-file-2 (vector 0 0 0 0 0 0 0 0 0 0 0 0 0)))

        (define metadata1 '((mime-type . "text/plain")
                            (name . "test 1.txt")))

        (define metadata2 '((mime-type . "text/plain")
                            (name . "test 2.txt")
                            (title . "A second test file")))

        (define ae1 (make-archive-entry
                     file1-key
                     file1-reused?
                     metadata1))

        (test file1-key (archive-entry-key ae1))
        (test file1-reused? (archive-entry-key-reused? ae1))
        (test metadata1 (archive-entry-alist ae1))

        (define ae2 (make-archive-entry
                     file2-key
                     file2-reused?
                     metadata2))

        (test-define "Import file 1 into the archive"
                     import1-key
                     (tag-archive-import! a
                                          "archive"
                                          (list ae1)
                                          '((test-step . 1))
                                          test-job))

        (define (ae->list ae)
          (list (archive-entry-key ae)
                (archive-entry-alist ae)))

        (test
         "Search archive (step 1)"
         `((,file1-key ,metadata1))
         (map ae->list
              (search-archive
               a "archive"
               '(and
                 (= ($ mime-type) "text/plain")))))

        (test-define "Import file 2 into the archive"
                     import2-key
                     (tag-archive-import! a
                                          "archive"
                                          (list ae2)
                                          '((test-step . 2))
                                          test-job))

        (define (ae< ae1 ae2)
          (string<? (car ae1) (car ae2)))

        (test
         "Search archive (step 2)"
         (sort `((,file1-key ,metadata1)
                 (,file2-key ,metadata2)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive"
                     #t))
               ae<))

        (test "Searching for all properties..."
              '(mime-type name title)
              (list-archive-properties a "archive" #t))

        (test "Searching for no properties..."
              '()
              (list-archive-properties a "archive" #f))

        (test "Searching for some properties..."
              '(mime-type name)
              (list-archive-properties a "archive" '(= ($ name) "test 1.txt")))
        (test "Searching for mime types..."
              '("text/plain")
              (list-archive-property-values a "archive" #t 'mime-type))

        (test "Searching for names..."
              '("test 1.txt" "test 2.txt")
              (list-archive-property-values a "archive" #t 'name))

        (test "Searching for titles..."
              '("A second test file")
              (list-archive-property-values a "archive" #t 'title))

        (test "Searching for titles where there are none..."
              '()
              (list-archive-property-values a "archive" '(= ($ name) "test 1.txt") 'title))

        ;; Update file2 on a new tag

        (define metadata3 '((mime-type . "text/plain")
                            (name . "test 2a.txt")))

        (define ae3 (make-archive-entry
                     file2-key
                     #t                 ; Definitely reused now!
                     metadata3))

        (test-define "Import file 2 into the second archive"
                     import3-key
                     (tag-archive-import! a
                                          "archive2"
                                          (list ae3)
                                          '((test-step . 3))
                                          test-job))


        (test
         "Search archive (step 3)"
         `((,file2-key ,metadata3))
         (map ae->list
              (search-archive
               a "archive2"
               '(and
                 (= ($ mime-type) "text/plain")))))

        (test "Searching for all properties..."
              '(mime-type name)
              (list-archive-properties a "archive2" #t))

        (test "Searching for no properties..."
              '()
              (list-archive-properties a "archive2" #f))

        (test "Searching for some properties..."
              '(mime-type name)
              (list-archive-properties a "archive2" '(= ($ name) "test 2a.txt")))

        (test "Cross-archive property leakage..."
              '()
              (list-archive-properties a "archive2" '(= ($ name) "test 1.txt")))

        (test "Searching for mime types..."
              '("text/plain")
              (list-archive-property-values a "archive2" #t 'mime-type))

        (test "Searching for names..."
              '("test 2a.txt")
              (list-archive-property-values a "archive2" #t 'name))

        (test "Searching for titles..."
              '()
              (list-archive-property-values a "archive2" #t 'title))

        (test "Merge archives" (void)
              (vault-merge-tags! a "archive3"
                                        ; archive2 is last, to override
                                        ; archive on the shared file2 key
                                   '("archive" "archive2")))

        (test
         "Search archive (step 4)"
         (sort `((,file1-key ,metadata1)
                 (,file2-key ,metadata3)) ae<) ; New metadata for file2
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     #t))
               ae<))

        (test "Searching for all properties..."
              '(mime-type name)         ; No titles any more
              (list-archive-properties a "archive3" #t))

        (test "Searching for mime types..."
              '("text/plain")
              (list-archive-property-values a "archive3" #t 'mime-type))

        (test "Searching for names..."
              '("test 1.txt" "test 2a.txt")
              (list-archive-property-values a "archive3" #t 'name))

        (test "Searching for titles..."
              '()
              (list-archive-property-values a "archive3" #t 'title))

        ;; Complex searches

        (test
         "Search archive (complex) - by import prop"
         (sort `((,file1-key ,metadata1)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     '(= ($import test-step) 1)))
               ae<))

        (test
         "Search archive (complex) - by key"
         (sort `((,file2-key ,metadata3)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     `(= key ,file2-key)))
               ae<))

        (test
         "Search archive (complex) - not key"
         (sort `((,file1-key ,metadata1)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     `(not (= key ,file2-key))))
               ae<))

        (test
         "Search archive (complex) - or"
         (sort `((,file2-key ,metadata3)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     `(or (= key ,file2-key)
                          (= ($ nonexistant) "fish"))))
               ae<))

        (test
         "Search archive (complex) - and"
         (sort `((,file1-key ,metadata1)) ae<)
         (sort (map ae->list
                    (search-archive
                     a "archive3"
                     `(and (= key ,file1-key)
                           (= ($import test-step) 1))))
               ae<))

        (test
         "Search archive (complex) - imports"
         (sort `((,file1-key ,metadata1 ,import1-key ((test-step . 1)))
                 (,file2-key ,metadata3 ,import3-key ((test-step . 3)))) ae<)
         (sort (map (lambda (ae)
                      (list (archive-entry-key ae)
                            (archive-entry-alist ae)
                            (archive-import-key (archive-entry-import ae))
                            (filter
                             (lambda (pe)
                               (eq? (car pe) 'test-step))
                             (archive-import-alist (archive-entry-import ae)))))
                    (search-archive
                     a "archive3"
                     #t))
               ae<))
        )))))

;; Handy wrapper to test all the "vault algorithms",
;; which are tests of the stuff above the very basic vault layer,
;; so unlikely to need re-testing for every backend
;; (although it's best to test it on a storage with small maximum
;; blocks so we have to use indirection, and
;; supporting unlinking so we can test that,
;; and with a file cache so we can test all of
;; that stuff).

(define (test-vault-algorithms a store-path)

  (test-group
   "Blocking on tag locks"
   (test "Make a tag" (void) (vault-set-tag! a (make-tag "TEST" 'test "TEST")))
   (test "Lock a tag" #t (vault-lock-tag! a "TEST"))
   (test-error "Lock a tag again" (vault-lock-tag! a "TEST"))
   (test "Unlock a tag" (void) (vault-unlock-tag! a "TEST"))
   (test "Remove the tag" (void) (vault-remove-tag! a "TEST")))

  (test-vault-streams a store-path)
  (test-vault-files a store-path)
  (test-vault-snapshots a store-path)
  (test-vault-archives a store-path)

  ;; FIXME: .ugarit / global directory rules tests

  ;; FIXME: Progress callback tests

  ;; FIXME: Unlinked snapshots

  ;;        Unlink a few of the snapshots.  Test that
  ;;        fold-snapshot-history still works.  Test that attempting
  ;;        to extract missing snapshots errors as expected.  Test
  ;;        that all the other snapshots extract OK.

  ;; FIXME: Missing blocks

  ;;        Unlink a single block from the large multi-block file, and
  ;;        attempt to extract affected snapshots. Check we get the
  ;;        error message, and a file with gaps in.

  ;;        Unlink an entire file and attempt to extract affected
  ;;        snapshots. Check for error message and missing file.

  ;;        Unlink a directory block and, again, extract and check for
  ;;        errors and missing subdirectory.
  (void)
  )

;; Actual Tests

(if (directory? "./tmp")
    (delete-directory "./tmp" #t))
(create-directory "./tmp")

(test-group
 "Test history DAG serialisation"

 ;; A test graph for fold-history** is represented
 ;; as a list of nodes, each of which is a three-element list:
 ;; key, mtime, list of parent keys.

 (define ((get-parents graph) key)
   (third (assoc key graph)))

 (define ((get-props graph) key)
   (list (cons 'mtime (second (assoc key graph)))))

 (define (kons key name props knil)
   (cons (cons key name) knil))

 (define (test-fold name graph root-key expected)
   (test name
         expected
         (reverse ;; Reverse, as fold results are back to front
          (map (lambda (element)
                 ; Each kons-result element is (key . name)
                 ; The names are nasty timestamps,
                 ; so let's make it (key . suffix)
                 ; as that's what the expected results are.
                 ; the magic number to string-drop
                 ; is the length of "YYYY-MM-DD HH:MM:SS"
                 (let ((key (car element))
                       (name (cdr element)))
                   (cons key (string-drop name 19)))
                 )
           (fold-history**
            (get-parents graph)
            (get-props graph)
            root-key
            kons
            '())))))

 ;; Simple linear cases

 (test-fold "Single Element"
            '(("a" 1 ())) "a" '(("a" . "")))

 (test-fold "Two elements, ignore 1"
            '(("a" 1 ())
              ("b" 2 ("a"))) "a" '(("a" . "")))

 (test-fold "Two elements"
            '(("a" 1 ())
              ("b" 2 ("a"))) "b" '(("b" . "")
                                   ("a" . "")))

 (test-fold "Three elements"
            '(("a" 1 ())
              ("b" 2 ("a"))
              ("c" 3 ("b"))) "c" '(("c" . "")
                                   ("b" . "")
                                   ("a" . "")))

 ;; Merges

 (test-fold "Simple merge"
            '(("a" 1 ())
              ("b" 2 ())
              ("c" 3 ("a" "b"))) "c" '(("c" . "") 
                                       ("b" . "-1")
                                       ("a" . "")))

 (test-fold "Simple merge, out of order"
            '(("a" 1 ())
              ("b" 2 ())
              ("c" 3 ("b" "a"))) "c" '(("c" . "") 
                                       ("b" . "-1")
                                       ("a" . "")))

 (test-fold "Three-way merge"
            '(("a" 1 ())
              ("b" 2 ())
              ("c" 3 ())
              ("d" 4 ("a" "b" "c"))) "d" '(("d" . "")
                                           ("c" . "-1")
                                           ("b" . "-2")
                                           ("a" . "")))

 (test-fold "Long three-way merge"
            '(("a" 1 ())
              ("b" 2 ())
              ("c" 3 ())
              ("d" 4 ("a"))
              ("e" 5 ("b"))
              ("f" 6 ("c"))
              ("g" 7 ("f" "d" "e")))
            "g"
            '(("g" . "")
              ("f" . "-1")
              ("e" . "-2")
              ("d" . "-3")
              ("c" . "-1")
              ("b" . "-2")
              ("a" . "")))

 ;; Forks
 ;; (which have to merge again to create a SINGLE history)

 (test-fold "Simple fork/merge"
            '(("a" 1 ())
              ("b" 2 ("a"))
              ("c" 3 ("a"))
              ("d" 4 ("b" "c"))) "d" '(("d" . "")
                                       ("c" . "-1")
                                       ("b" . "-2")
                                       ("a" . "")))

 ;; a->(b,c), b->(d,e) (c,e)->f, (d,f)->g
 (test-fold "Double fork/merge"
            '(("a" 1 ())
              ("b" 2 ("a"))
              ("c" 3 ("a"))
              ("d" 4 ("b"))
              ("e" 5 ("b"))
              ("f" 6 ("e" "c"))
              ("g" 7 ("d" "f")))
            "g"
            '(("g" . "")
              ("f" . "-1")
              ("e" . "-2")
              ("d" . "-3")
              ("c" . "-4")
              ("b" . "-5")
              ("a" . "")))

 ;; a->(b,c) b->d c->e a->f (d,e,f)->g
 (test-fold "Fat fork/merge"
            '(("a" 1 ())
              ("b" 2 ("a"))
              ("c" 3 ("a"))
              ("d" 4 ("b"))
              ("e" 5 ("c"))
              ("f" 6 ("a"))
              ("g" 7 ("e" "d" "f")))
            "g"
            '(("g" . "") 
              ("f" . "-1")
              ("e" . "-2")
              ("d" . "-3")
              ("c" . "-2")
              ("b" . "-3")
              ("a" . ""))))

(test-group
 "Test different backends"

 (test-group
  "Filesystem backend"
  (create-directory "./tmp/be1")
  (test-define "Open storage" be (import-storage "backend-fs fs ./tmp/be1"))
  (test-backend be))

 (test-group
  "Splitlog backend"
  (create-directory "./tmp/be3")
  (test-define "Open storage" be (import-storage "backend-fs splitlog ./tmp/be3 ./tmp/be3/metadata"))
  (test "Set max file size to 1024" '((result . "Done")) ((storage-admin! be) '(set-max-logfile-size! 1024)))
  (test-backend be))

 (test-group
  "Limited cached splitlog backend"
  (create-directory "./tmp/be4")
  (test-define "Open storage" be (import-storage "backend-cache ./tmp/be4-cache \"backend-fs splitlog ./tmp/be4 ./tmp/be4/metadata\""))
  (test-define "Wrap in block-limiter" lbe (backend-limit-block-size be 1024))
  (test-backend lbe))

 (test-group
  "sqlite backend"
  (test-define "Open storage" bex (import-storage "backend-log /tmp/foo \"backend-sqlite ./tmp/be4a.vault\""))
  (test-backend bex)))

(test-group
 "Smoke test vault operations on various backends and configurations"
 (define expected-tiger-hash "03cf987c795c71b791c75ade190d42776af9b9a7f19a17ee")

 (test-group
  "Filesystem backend vault"
  (create-directory "./tmp/be5")
  (test-define "Open vault" be (open-vault '((storage "backend-fs fs ./tmp/be5"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "Filesystem backend vault plus file cache"
  (create-directory "./tmp/be6")
  (test-define "Open vault" be (open-vault '((storage "backend-fs fs ./tmp/be6") (file-cache "./tmp/be6-file-cache"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "Splitlog backend vault"
  (create-directory "./tmp/be7")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be7 ./tmp/be7/metadata"))))
  (test "Set max file size to 1024" '((result . "Done")) (vault-admin! be '(set-max-logfile-size! 1024)))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "Splitlog backend vault plus file cache"
  (create-directory "./tmp/be8")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be8 ./tmp/be8/metadata") (file-cache "./tmp/be8-file-cache"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "sqlite backend plus file cache"
  (test-define "Open vault" be (open-vault '((storage  "backend-sqlite ./tmp/be8a.vault") (file-cache "./tmp/be8a-file-cache"))))
  (test-vault-basics be "./tmp/be8a.vault" expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "tiger hash with a key"
  (create-directory "./tmp/be8a")
  (test-define "Open vault" be
               (open-vault '((storage  "backend-log -v ./tmp/be8a.log \"backend-fs splitlog ./tmp/be8a ./tmp/be8a/metadata\"") (file-cache "./tmp/be8a-file-cache") (hash tiger "Hello World"))))
  (test-vault-basics be #f "3028f8f02fb9e8035fdeac2eec9113a0658d75cb96adce00")
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "sha256 hash"
  (create-directory "./tmp/be9")
  (test-define "Open vault" be
               (open-vault '((storage  "backend-log -v ./tmp/be9.log \"backend-fs splitlog ./tmp/be9 ./tmp/be9/metadata\"") (file-cache "./tmp/be9-file-cache") (hash sha256))))
  (test-vault-basics be #f "6e06213fca35e107ec7c58a4b1c1eb6624b933cc0e122dc44b340f9df35ccf03")
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "sha256 hash with a key"
  (create-directory "./tmp/be10")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be10 ./tmp/be10/metadata") (file-cache "./tmp/be10-file-cache") (hash sha256 "Hello World"))))
  (test-vault-basics be #f "dc191eba76659b9f777f6eac6a42dae1e63498e8358671b195603c7baf9c2c47")
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes passphrase 32"
  (create-directory "./tmp/be11")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be11 ./tmp/be11/metadata") (file-cache "./tmp/be11-file-cache") (encryption aes (32 "Hello, world")))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes passphrase 24"
  (create-directory "./tmp/be12")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be12 ./tmp/be12/metadata") (file-cache "./tmp/be12-file-cache") (encryption aes (24 "Hello, world")))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes passphrase 16"
  (create-directory "./tmp/be13")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be13 ./tmp/be13/metadata") (file-cache "./tmp/be13-file-cache") (encryption aes (16 "Hello, world")))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes256"
  (create-directory "./tmp/be14")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be14 ./tmp/be14/metadata") (file-cache "./tmp/be14-file-cache") (encryption aes "00112233445566778899AABBCCDDEEFF"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes192"
  (create-directory "./tmp/be15")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be15 ./tmp/be15/metadata") (file-cache "./tmp/be15-file-cache") (encryption aes "00112233445566778899AABBCCDDEEFF0011223344556677"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "aes128"
  (create-directory "./tmp/be16")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be16 ./tmp/be16/metadata") (file-cache "./tmp/be16-file-cache") (encryption aes "00112233445566778899AABBCCDDEEFF00112233445566778899AABBCCDDEEFF"))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be)))

 (test-group
  "deflate"
  (create-directory "./tmp/be17")
  (test-define "Open vault" be (open-vault '((storage  "backend-fs splitlog ./tmp/be17 ./tmp/be17/metadata") (file-cache "./tmp/be17-file-cache") (compression deflate))))
  (test-vault-basics be #f expected-tiger-hash)
  (test "Close vault" (void) (vault-close! be))))

(test-group "Bad vault tests"
  (test-group "Set up a hashed, encrypted, vault"
              (create-directory "./tmp/be18")
              (test-define "Open vault" be
                           (open-vault '((storage  "backend-fs splitlog ./tmp/be18 ./tmp/be18/metadata")
                                         (file-cache "./tmp/be18-file-cache")
                                         (encryption aes "00112233445566778899AABBCCDDEEFF00112233445566778899AABBCCDDEEFF")
                                         (hash sha256 "Hello World"))))

              (test "Close vault" (void) (vault-close! be)))

  (test-error "Try to open it with the wrong hash"
              (open-vault '((storage  "backend-fs splitlog ./tmp/be18 ./tmp/be18/metadata")
                            (file-cache "./tmp/be18-file-cache")
                            (encryption aes "00112233445566778899AABBCCDDEEFF00112233445566778899AABBCCDDEEFF")
                            (hash sha256 "Goodbye World"))))

  (test-error "Try to open it with the wrong encryption key"
              (open-vault '((storage  "backend-fs splitlog ./tmp/be18 ./tmp/be18/metadata")
                            (file-cache "./tmp/be18-file-cache")
                            (encryption aes "00112233445566778899AABBCCDDEEFF00112233445566778899AABBCCDDEEFD")
                            (hash sha256 "Hello World"))))

  (test-error "Try to open it with the wrong encryption key and hash"
              (open-vault '((storage  "backend-fs splitlog ./tmp/be18 ./tmp/be18/metadata")
                            (file-cache "./tmp/be18-file-cache")
                            (encryption aes "00112233445566778899AABBCCDDEEFF00112233445566778899AABBCCDDEEFD")
                            (hash sha256 "Goodbye World")))))

(test-group "Vault algorithms"
            (test-define "Open vault" be (open-vault '((storage  "backend-sqlite ./tmp/be19.vault") (file-cache "./tmp/be19-file-cache"))))

            ;; This helps exercise the algorithms, by forcing the generation of indirect blocks in some tests
            (test "Set block size to 1024" '((result . "Done")) (vault-admin! be '(set-block-size! 1024)))
            (test-vault-algorithms be "./tmp/be19.vault")
            (test "Close vault" (void) (vault-close! be)))

(test-group "Command line interface tests"
 (let* ((vault-path "./tmp/cli.vault")
        (cache-path "./tmp/cli-cache")
        (data-dir "./tmp/cli-data")
        (extract1-dir "./tmp/cli-extract1")
        (extract2-dir "./tmp/cli-extract2")
        (ugarit-conf-file "./tmp/cli.conf")
        (tp (lambda (file) (make-pathname data-dir file))))

  (create-directory data-dir)
  (create-directory extract1-dir)
  (create-directory extract2-dir)
  ;; FIXME: Put in some .ugarit files, too!
  (with-output-to-file
      (tp "plain-file")
      (lambda ()
        (printf "Hello, world!\n")))

  (with-output-to-file
      ugarit-conf-file
      (lambda ()
        (write `(storage
                 ,(string-append "backend-sqlite "
                                 vault-path)))
        (newline)
        (write `(file-cache ,cache-path))
        (newline)))

  (test "Snapshot" 0
        (system (string-append
                 "ugarit snapshot " ugarit-conf-file
                 " test " data-dir)))

  (define (test-extract-results dir)
    (test-assert "Top-level directory exists" (directory-exists? dir))
    (let ((plain-file (string-append dir "/plain-file")))
      (test-assert "Test file exists"
                   (file-exists? plain-file))
      (test "Test file is correct" "Hello, world!"
            (with-input-from-file
                plain-file read-line))))

  (test-group "Extract"
     (test "Extract" 0
           (system (string-append
                    "ugarit extract " ugarit-conf-file
                    " test/current/contents " extract1-dir)))
     (test-extract-results (make-pathname extract1-dir "contents")))

 ;; FIXME
 ;; Restore it
 ;; Fork it and restore the fork to check it worked

  ;; Import into an archive
  ;; Search an archive
  ;; Extract from an archive (dir -> disk)
  ;; Extract from an archive (stream to stdout)

 ;; Test explore mode somehow?
))

(printf "Final count of failures: ~a\n" (test-failure-count))

(test-exit)
